/****************************************************************************** 
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2010. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
 // Summary:
 //    An abstract implementation of a virtual content table provider. This provides
 //    an base implementation that handles common processing for virtual content clients.
 //    The concrete clients need to supply the implementations of methods that actually
 //    retrieve the virtual contents. 
 //
dojo.provide("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.DefaultVirtualContentProvider");

dojo.require("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableContentProvider");
dojo.require("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableConstants");

(function() {

var TableContentProvider = com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableContentProvider;
var TableConstants = com.ibm.asq.common.web.ui.internal.widgets.tableViewer.TableConstants; 

dojo.declare("com.ibm.asq.common.web.ui.internal.widgets.tableViewer.DefaultVirtualContentProvider", TableContentProvider,
{
    hostTableViewer: null,
    isVirtual: true,
    totalContentCount: -1,
    totalPageCount: 0,
    currentPageIndex: 0,
    pageSize: TableConstants.MIN_PAGE_SIZE,
    filterText: null,
    
    constructor: function(args) 
    {
        dojo.mixin(this, args);     
    },
    
    // Summary:
    //    Abstract method for the virtual content provider to call to fetch a virtual page.
    //    Client needs to provide an implementation to actually fetches a page's content.
    //
    vcFetchVirtualPage: function(pageIndex, pageSize)
    {
    },

    // Summary:
    //    Abstract method for the virtual content provider to call to run a new query.
    //    Client needs to provide an implementation to actually perform the query.
    //
    vcRunQuery: function(pageSize)
    {
    },

    // Summary:
    //    Returns whether the content is virtual, meaning that only partial content is
    //    loaded locally on the client. If this method returns true, the table viewer will delegate
    //    all content related processing to the provider using the following methods prefixed by 'vc'.
    //
    //    Note: Only clients with virtual content need to implement this method and the 'vc' methods.
    //    
    isVirtualContent: function()
    {
        // returns true if the content is virtual.
        return this.isVirtual;
    },
    
    // Summary:
    //    Returns the total virtual content count. If the content's count is unknown, the client
    //    should return the maximum page size. This allows the table to create all of the selectable
    //    page sizes for the user. 
    //
    vcGetContentItemCount: function()
    {    
        return this.totalContentCount;
    },
    
    // Summary:
    //    Set and change the visible page size of the content. 
    //
    vcSetPageSize: function(newPageSize)
    {    
        this.pageSize = newPageSize;
        
        // set the new page size. We'll have to
        // reset to the first page to fetch from
        // the top
        this.vcPerformFirstPage();
    },
     
    // Summary:
    //    Returns the total virtual content's page count.
    //
    vcGetTotalPageCount: function()
    { 
        return this.totalPageCount;   
    },
    
    // Summary:
    //    Returns the current page's index value 0 based, i.e., 0 to total page count - 1
    //
    vcGetCurrentPageIndex: function()
    {    
        return this.currentPageIndex;
    },
    
    // Summary:
    //    Returns true if the virtual content has next page. 
    //
    vcGetHasNextPage: function()
    {
        return this.currentPageIndex < this.totalPageCount - 1;
    },

    // Summary:
    //    Returns true if the virtual content has previous page. 
    //
    vcGetHasPreviousPage: function()
    {
        return this.currentPageIndex > 0;
    },
         
    // Summary:
    //    Display the first page of the virtual content
    //
    vcPerformFirstPage:function()
    {
        this.currentPageIndex = 0;
        this.vcFetchVirtualPage(this.currentPageIndex, this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Display the last page of the virtual content
    //
    vcPerformLastPage: function()
    {
        this.currentPageIndex = this.totalPageCount - 1;
        this.vcFetchVirtualPage(this.currentPageIndex, this.pageSize);
        this.preLoadPageContent();
    },

    // Summary:
    //    Display the previous page of the virtual content
    //
    vcPerformPreviousPage: function()
    {
        this.currentPageIndex -= 1;
        this.vcFetchVirtualPage(this.currentPageIndex, this.pageSize);
        this.preLoadPageContent();
    },

    // Summary:
    //    Display the next page of the virtual content
    //
    vcPerformNextPage: function()
    {
        this.currentPageIndex += 1;
        this.vcFetchVirtualPage(this.currentPageIndex, this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Display the specified page of the virtual content
    //
    vcPerformIndexPage: function(pageIndex)
    {
        if (pageIndex != this.currentPageIndex) {
            this.currentPageIndex = pageIndex;
            this.vcFetchVirtualPage(this.currentPageIndex, this.pageSize);
            this.preLoadPageContent();
        }
    },
    
    // Summary:
    //    Single sort the virtual content on the specified column index
    //
    vcPerformSortBy: function(colIndex, sortDir)
    {
        this.vcRunQuery(this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Group the virtual content on the specified column index
    //
    vcPerformGroupBy: function(colIndex)
    {
        this.vcRunQuery(this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Filter content by free form text
    //
    vcPerformFilterBy: function(filterText)
    {
        this.filterText = filterText;
        this.vcRunQuery(this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Returns the current free form text filter.
    //
    getTextFilter: function()
    {
        return this.filterText;    
    },
    
    // Summary:
    //    Muiti-column sort the virtual content on the specified columns.
    //    SortColumns is expected to be array holding the sort columns in order, 0 = 1st, 1 = 2nd, etc.
    //
    //     [ 
    //         {'col': column position, 'dir': direction},
    //         ...
    //     ]
    //
    vcPerformMultiColumnSort: function(sortColumns)
    {
        this.vcRunQuery(this.pageSize);
        this.preLoadPageContent();
    },
    
    // Summary:
    //    Before fetching a server page, clear the current contents and set the table
    //    to loading.
    preLoadPageContent: function()
    {
        this.hostTableViewer.setVirtualContent([]);
        this.hostTableViewer.showLoadingProgress();        
    },

    // Summary:
    //    Before fetching a server page, clear the current contents and set the table
    //    to loading.
    loadPageContent: function(content, msg)
    {        
        this.hostTableViewer.endLoadingProgress();
        this.hostTableViewer.setVirtualContent(content || []); 
        
        if(!content && msg) {
            this.hostTableViewer.showContentError(msg);
        }
    },
    
    // Summary:
    //    Compose the server query arguments for table's sort and group settings.
    //
    buildGroupSortServerArgs: function()
    {
        var hostTableViewer = this.hostTableViewer;
        var sortArgs = [];
        var groupByCol = hostTableViewer.getGroupByColumn();
        
        // apply grouping, it's just a first level sort.
        if (groupByCol > -1) {
            var sortParam = this.getColumnIdName(groupByCol) + ":ASCENDING";
            sortArgs.push(sortParam);
        }
        // apply sorting
        if (hostTableViewer.allowMultiSort && hostTableViewer.sortColumns) {
            var sorts = hostTableViewer.sortColumns;
            
            for (var i = 0; i < sorts.length; i++) {
                var aSort = sorts[i];
                var sortParam = this.getColumnIdName(aSort.col) + ":" + ((aSort.dir == TableConstants.SORT_DIR_ASCEND) ? "ASCENDING" : "DESCENDING");
                sortArgs.push(sortParam);
            }
        }
        else {
            // handle single column sort case
            var sortByCol = hostTableViewer.getSortByColumn();
            if (sortByCol > -1) {
                var sortDir = hostTableViewer.getSortDirection();
                sortArgs.push(this.getColumnIdName(sortByCol) +  ":" + ((sortDir == TableConstants.SORT_DIR_ASCEND) ? "ASCENDING" : "DESCENDING"));
            }
        }
        return sortArgs;
    }
});

})();
