/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008, 2010. All Rights Reserved.
 *  
 * U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 ******************************************************************************/
//-*- tab-width: 4 -*-
package com.ibm.rational.test.lm.spi.bf;

import java.io.IOException;
import java.sql.Timestamp;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.buildforge.services.client.api.APIClientConnection;
import com.buildforge.services.client.dbo.Build;
import com.buildforge.services.client.dbo.Environment;
import com.buildforge.services.client.dbo.EnvironmentEntry;
import com.buildforge.services.client.dbo.Log;
import com.buildforge.services.client.dbo.Project;
import com.buildforge.services.client.dbo.Result;
import com.buildforge.services.client.dbo.Selector;
import com.buildforge.services.client.dbo.SelectorProperty;
import com.buildforge.services.client.dbo.Server;
import com.buildforge.services.client.dbo.ServerAuth;
import com.buildforge.services.common.ServiceException;
import com.buildforge.services.common.dbo.BuildDBO;
import com.buildforge.services.common.dbo.BuildDBO.State;
import com.ibm.rational.test.lm.spi.bf.common.BFAPIClientConnectionException;
import com.ibm.rational.test.lm.spi.bf.common.BFAPIException;
import com.ibm.rational.test.lm.spi.bf.common.BFConstants;
import com.ibm.rational.test.lm.spi.bf.common.TopologyParser;
import com.ibm.rational.test.lm.spi.bf.common.messages.Messages;
import com.ibm.rational.test.lm.spi.bf.internal.APIExceptionHandle;
import com.ibm.rational.test.lm.spi.bf.internal.BFAPIClientConnection;
import com.ibm.rational.test.lm.spi.bf.internal.BuildForgeArtifacts;
import com.ibm.rational.test.lm.spi.bf.utils.CreateTopology;
import com.ibm.rational.test.lm.spi.common.SPIException;
import com.ibm.rational.test.lm.spi.common.SPIResourceNotFoundException;

/**
 * Build Forge Job Controller will control all the RTLM related Build Forge
 * jobs: get the job list, run a particular job on a particular machine, get the
 * job result,etc.
 * 
 */
public class AutomationService {

	// singleton
	public static final AutomationService INSTANCE = new AutomationService();

	/**
	 * Cancel a running/pending job through the job Id. If the job's status is
	 * COMPLETE or WAITING, the function will throw SPI Exception
	 * 
	 * @param jobId
	 * @param provider_server_name
	 * @param provider_server_port
	 * @throws SPIException
	 * @throws InterruptedException
	 */
	@SuppressWarnings("static-access")
	public void cancelJob(String jobId, String provider_server_name,
			int provider_server_port) throws SPIException {
		BFAPIClientConnection bfConn = null;
		try {
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);

			Build job = getBuildById(bfConn, jobId);
			int state = Integer.valueOf(getJobStatusbyJobId(jobId,
					bfConn));
			// Waiting the job to waiting and running, then we can cancel it.
			while ( state != BFConstants.WAITING
					& state != BFConstants.RUNNING) {
				
				if (Integer.valueOf(getJobStatusbyJobId(jobId,
						bfConn)) != BFConstants.COMPLETE) {
					try {
						Thread.currentThread().sleep(1);
					} catch (InterruptedException e) {

					}
				} else {
					throw new SPIException("Job has finished ,Can't Cancel");//$NON-NLS-1$
				}
			}
			
			job.cancel();

		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		}
		finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
//					throw new SPIException(
//							"BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
	}

	/**
	 * Fire build forge 's job definition through the jobDefId
	 * 
	 * @param jobDefId
	 * @param serverId
	 * @param parameterNameandValueMap
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	private String executeJobbyId(String jobDefId, String serverId,
			HashMap<String, String> parameterNameandValueMap,
			String provider_server_name, int provider_server_port,
			String userName, String password, String rlmAgentDirectory)
			throws SPIException {
		BFAPIClientConnection bfConn = null;
		try {
			APIClientConnection connection = null;
			// Get the connection of Build Forge
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);
			connection = bfConn.getConnection();

			Project jobDef = null;
			Server server = null;
			try {
				jobDef = Project.findById(connection, Integer.valueOf(jobDefId));
				server = Server.findById(connection, Integer.valueOf(serverId));
			} catch (Exception e) {
				// TODO log it?
			}
			if (jobDef == null) {
				throw new SPIResourceNotFoundException("The job is not found");
					
			}
			if (server == null) {
				throw new SPIResourceNotFoundException("The machine is not found");
			}

			BuildForgeArtifacts artifacts = new BuildForgeArtifacts();
			// Set the selector id to name of the server
			String serverName = server.getName();
			String selectorId = serverName;
			Selector selector = null;
			try {
				selector = Selector.findById(connection, selectorId);
			} catch (Exception e) {
			}
			if (selector != null) {
				Collection<SelectorProperty> collection = selector
						.getProperties();
				SelectorProperty[] selProperties = collection
						.toArray(new SelectorProperty[collection.size()]);
				if (!selProperties[0].getPropertyName().equals(BFConstants.SELECTOR_PROPERTY_NAME)
						|| !selProperties[0].getPropertyValue().equals(selectorId)) {
					artifacts.updateRlmSelector(selProperties[0], selectorId);
				}
			} else {
				artifacts.createRlmSelector(connection, selectorId);
			}

			String authId = server.getAuthId();
			ServerAuth serverAuth = null;
			// First we'll try to use the server authentication that is
			// already associated with the server.
			if (authId != null && authId.length() != 0) {
				try {
					serverAuth = ServerAuth.findById(connection, authId);
				} catch (Exception e) {
				}
			}
			// If there is no auth associated then we'll look one up based
			// on the name RQM uses when creating one.
			if (serverAuth == null) {
				serverAuth = ServerAuth.findById(connection,
						BFConstants.serverAuthIdPrefix + serverName);
			}
			if (serverAuth != null) {
				List<EnvironmentEntry> params = setEnvironmentParamToProject(
						bfConn, jobDef, parameterNameandValueMap);
				Build build = Build.fire(connection, jobDef.getProjectId(),
						selectorId, jobDef.getBuildClass(), false, params);
				return getJobDetails(build, jobDef);
			} else {
				throw new SPIException(
						"Can't get the authentication of the server [" + serverName + "] in Build Forge");//$NON-NLS-1$
			}
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
					// throw new SPIException(
					// "BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
		return null;
	}

	/**
	 * Parse xml string to get build forge 's project id and fire the job
	 * definition to run
	 * 
	 * @param xmlBody
	 * @param hostname
	 * @param ipAddress
	 * @param params
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	public String executeJobDef(String xmlBody, String provider_server_name,
			int provider_server_port) throws SPIException {

		TopologyParser topParser = new TopologyParser();
		String[] jobDefIds = topParser.parseToJobDefId(xmlBody);
		String[] serverIds = topParser.parseToServerId(xmlBody);
		HashMap<String, String> params = topParser.parseToJobParams(xmlBody);
		String xml = "";//$NON-NLS-1$
		String jobDefId = jobDefIds[0];
		String targetServerId = serverIds[0];
		// TODO Need to parse the username, password, and install directory in
		// the xml string.
		xml = executeJobbyId(jobDefId, targetServerId, params,
				provider_server_name, provider_server_port, null, null, null);
		return xml;

	}

	/**
	 * Get build forge's job definition by project id
	 * 
	 * @param jobDefID
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	public String getJobDef(String jobDefID, String provider_server_name,
			int provider_server_port) throws SPIException {

		BFAPIClientConnection bfConn = null;

		try {
			// Connect to the Build Forge
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);

			HashMap<Project, HashMap<String, String>> projectToParamsMap = new HashMap<Project, HashMap<String, String>>();
			Project[] jobDefs = new Project[1];
			jobDefs[0] = Project.findById(bfConn.getConnection(), Integer
					.valueOf(jobDefID));
			if (jobDefs[0] == null) {
				return null;
			}
			HashMap<String, String> parms = this.listProjectEnvironmentInfo(
					bfConn, jobDefs[0]);
			projectToParamsMap.put(jobDefs[0], parms);

			// Get Job Definition Java Bean from BF Project
			String jobDef = jobDefToXML(projectToParamsMap);
			return jobDef;

		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
					// throw new SPIException(
					// "BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
		return null;
	}

	/**
	 * Get build forge job's detail information through build id
	 * 
	 * @param jobId
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	public String getJobDetailsByJobId(String jobId,
			String provider_server_name, int provider_server_port)
			throws SPIException {
		BFAPIClientConnection bfConn = null;
		try {
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);

			// Find the build by Job Id
			Build job = getBuildById(bfConn, jobId);
			if (job == null) {
				return null;
			}
			Project jobDef = getProjectById(bfConn, String.valueOf(job
					.getTargetId()));
			Build[] buildArray = new Build[1];
			buildArray[0] = job;

			return getJobDetails(job, jobDef);

		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
					// throw new SPIException(
					// "BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
		return null;
	}

	private String getJobDetails(Build build, Project project){
		Build[] buildArray = new Build[1];
		buildArray[0] = build;

		Project[] projectArray = new Project[1];
		projectArray[0] = project;
		return jobResultToXML(buildArray, projectArray);
	}
	
	/**
	 * Get job log from build
	 * 
	 * @param job
	 * @return
	 * @throws IOException
	 * @throws ServiceException
	 */
	public String getJobLog(Build job, String[] logTypes) {

		Iterator<Result> keyItr;
		try {
			keyItr = job.getResults().iterator();
			StringBuffer buildLogBuf = new StringBuffer();
			while (keyItr.hasNext()) {
				Result buildResult = keyItr.next();
				List<Log> logs = buildResult.getLogs();
				int stepId = buildResult.getResultStepId();
				if (logs != null) {
					for (Log log : logs) {
						for (String logType : logTypes) {
							if (logType.equals(log.getType())) {
								buildLogBuf
										.append(stepId
												+ "  "
												+ log.getLineId()
												+ "  "
												+ String
														.valueOf(new Timestamp(
																Integer
																		.valueOf(log
																				.getStamp()) * 1000L))
												+ "  " + log.getType() + " "
												+ log.getMessageText() + "\r\n");
							}
						}
					}
				}
			}
            // If there are no log results and no notes, we either try to determine if
            // a) there are no log events yet because the job is running
            // b) the job could not run because there are no servers (should be a Note
            // for BF 7.1 but will be nothing for BF 7.0.2
            if (buildLogBuf.length() == 0) {
                if (job.getState().equals(BuildDBO.State.COMPLETED)) {
                    buildLogBuf.append(Messages.NO_SERVERS);
                } else {
                    buildLogBuf.append(Messages.NO_LOG);
                }
            }			
			return buildLogBuf.toString();
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		} catch (ServiceException e) {
			return null;
		}

	}

	/**
	 * Get job log information by job id(build id)
	 * 
	 * @param jobId
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	public String getJobLogByJobId(String jobId, String provider_server_name,
			int provider_server_port) throws SPIException {
		BFAPIClientConnection bfConn = null;
		try {
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);
			// Find the build by Job Id
			Build job = getBuildById(bfConn, jobId);
			if (job == null) {
				return null;
			}
			return jobLogToXML(job);
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
//					throw new SPIException(
//							"BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
		return null;

	}

	/**
	 * Get job status by build id
	 * 
	 * @param jobId
	 * @param connection TODO
	 * @return
	 * @throws SPIException
	 * @throws BFAPIClientConnectionException
	 */
	public String getJobStatusbyJobId(String jobId,
			BFAPIClientConnection connection)
			throws SPIException {
		BFAPIClientConnection bfConn = null;
		try {
			// Connect to the Build Forge

			Build job = getBuildById(connection, jobId);
			State state = job.getState();
			int status = BFConstants.NONE;

			// Use the status in Job Bean
			if ((BuildDBO.State.WAITING).equals(state)) {
				status = BFConstants.WAITING;
			}
			if (BuildDBO.State.COMPLETED.equals(state)) {
				status = BFConstants.COMPLETE;
			} else if (BuildDBO.State.RUNNING.equals(state)) {
				status = BFConstants.RUNNING;
			} else if (BuildDBO.State.WAITING.equals(state)) {
				status = BFConstants.WAITING;
			} else if (BuildDBO.State.LOCKED.equals(state)) {
				status = BFConstants.LOCKED;
			} else if (BuildDBO.State.PENDING.equals(state)) {
				status = BFConstants.PENDING;
			}
			return String.valueOf(status);
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
					// throw new SPIException(
					// "BuildForge APIClientConnectionException");//$NON-NLS-1$
				}
			}
		}
		return null;
	}

	/**
	 * List all job definition(project) id
	 * 
	 * @param provider_server_name
	 * @param provider_server_port
	 * @return
	 * @throws SPIException
	 */
	public String listAllJobDefIds(String provider_server_name,
			int provider_server_port) throws SPIException {
		BFAPIClientConnection bfConn = null;

		try {
			// Connect to the Build Forge
			bfConn = BFAPIClientConnection.getConnection(provider_server_name,
					provider_server_port);

			List<Project> allJobDefs = null;
			allJobDefs = Project.findAll(bfConn.getConnection());
			if (allJobDefs == null) {
				return null;
			}
			// List<Project> rtlmJobDefs = new ArrayList<Project>();
			HashMap<Project, HashMap<String, String>> projectToParamsMap = new HashMap<Project, HashMap<String, String>>();

			for (Project jobDef : allJobDefs) {
				// if (jobDef.getBuildClass()
				// .equals(BFConstants.rtlmJobBuildClass)) {
				// rtlmJobDefs.add(jobDef);

				HashMap<String, String> parms = this
						.listProjectEnvironmentInfo(bfConn, jobDef);
				projectToParamsMap.put(jobDef, parms);

				// }
			}

			if (!projectToParamsMap.isEmpty()) {
				return jobDefToXML(projectToParamsMap);
			} else
				// Return empty array if no suitable project.
				return null;
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		} finally {
			if (bfConn != null) {
				try {
					bfConn.close();
				} catch (BFAPIClientConnectionException e) {
					// throw new SPIException(
					// "BuildForge APIClientConnectionException"); //$NON-NLS-1$
				}
			}
		}
		return null;
	}

	/**
	 * Get the Build Forge's build object by buildId
	 * 
	 * @param conn
	 *            Build Forge connection
	 * @param buildId
	 *            Build Forge's build id
	 * @return Build Forge's build object
	 * @throws BFAPIClientConnectionException
	 * @throws ServiceException
	 * @throws IOException
	 * @throws SPIException
	 */
	private Build getBuildById(BFAPIClientConnection conn, String buildId)
			throws IOException, ServiceException,
			BFAPIClientConnectionException, SPIException {

		Build build = Build.findById(conn.getConnection(), Integer
				.valueOf(buildId));
		if (build == null) {
			throw new SPIException("Build can't be null."); //$NON-NLS-1$
		}
		return build;
	}

	/**
	 * Find Build Forge project by project id
	 * 
	 * @param conn
	 * @param jobDefId
	 * @return
	 * @throws RlmInternalException
	 * @throws JobDefinitionNotFoundException
	 * @throws ServiceException
	 * @throws IOException
	 * @throws BFAPIClientConnectionException
	 * @throws BFAPIException
	 */
	private Project getProjectById(BFAPIClientConnection conn, String jobDefId)
			throws IOException, ServiceException,
			BFAPIClientConnectionException, SPIException {
		Project job = null;

		job = Project.findById(conn.getConnection(), Integer.valueOf(jobDefId));

		if (job == null) {
			throw (new SPIException("Job can't be null.")); //$NON-NLS-1$
		}

		return job;
	}

	/**
	 * Convert job Definition to XML
	 * 
	 * @param projectWithParams
	 * @return a XML string of machine's detail information
	 */
	private String jobDefToXML(
			HashMap<Project, HashMap<String, String>> projectWithParams) {

		CreateTopology createTop = CreateTopology.getInstance();
		return createTop.createJobDefinitionTopology(projectWithParams);
	}

	/**
	 * Convert job result to XML
	 * 
	 * @param jobs
	 * @param jobDefs
	 * @return a XML string of machine's detail information
	 */
	private String jobResultToXML(Build[] jobs, Project[] jobDefs) {

		CreateTopology createTop = CreateTopology.getInstance();
		return createTop.createJobResultTopology(jobs, jobDefs);
	}

	private String jobLogToXML(Build job) {
		CreateTopology createTop = CreateTopology.getInstance();
		return createTop.createJobLogTopology(job);
	}

	/**
	 * List project related environment entry's name and Value
	 * 
	 * @param conn
	 * @param project
	 * @return hash map The key is parameter's name The value is parameter's
	 *         value
	 * @throws SPIException
	 */
	private HashMap<String, String> listProjectEnvironmentInfo(
			BFAPIClientConnection conn, Project project) throws SPIException {

		HashMap<String, String> parameterNameandValueMap = new HashMap<String, String>();
		try {
			int environmentId = project.getEnvironmentId();
			if (environmentId != 0) {
				Environment env = Environment.findById(conn.getConnection(),
						environmentId);
				List<EnvironmentEntry> envEntrys = env.getEntries();

				for (EnvironmentEntry entry : envEntrys) {
					String name = entry.getParameterName();
					String value = entry.getParameterValue();
					parameterNameandValueMap.put(name, value);
				}
			}
			return parameterNameandValueMap;
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		}
		return null;
	}

	/**
	 * set Environment Param To Project
	 * 
	 * @param project
	 * @param parameterNameandValueMap
	 * @return
	 * @throws SPIException
	 */
	private List<EnvironmentEntry> setEnvironmentParamToProject(
			BFAPIClientConnection conn, Project project,
			HashMap<String, String> parameterNameandValueMap)
			throws SPIException {
		try {
			int environmentId = project.getEnvironmentId();
			if (environmentId != 0) {
				Environment env = Environment.findById(conn.getConnection(),
						environmentId);
				Iterator<String> paramNameItr = parameterNameandValueMap
						.keySet().iterator();

				while (paramNameItr.hasNext()) {
					String paramName = paramNameItr.next();
					String paramValue = parameterNameandValueMap.get(paramName);
					if (env.getEntry(paramName) != null) {
						env.getEntry(paramName).setParameterValue(paramValue);
					} else {
						EnvironmentEntry newEntry = new EnvironmentEntry(conn
								.getConnection(), env);
						newEntry.setParameterName(paramName);
						newEntry.setParameterValue(paramValue);
						env.addEntry(newEntry);
					}
				}
				return env.getEntries();
			} else {
				Environment env = new Environment(conn.getConnection());
				Iterator<String> paramNameItr = parameterNameandValueMap
						.keySet().iterator();

				while (paramNameItr.hasNext()) {
					String paramName = paramNameItr.next();
					String paramValue = parameterNameandValueMap.get(paramName);
					EnvironmentEntry newEntry = new EnvironmentEntry(conn
							.getConnection(), env);
					newEntry.setParameterName(paramName);
					newEntry.setParameterValue(paramValue);
					env.addEntry(newEntry);
				}
				return env.getEntries();
			}
		} catch (Exception e) {
			APIExceptionHandle.handleAPIExceptions(e);
		}
		return null;
	}

}
